/*
   Copyright (c) 2012 LinkedIn Corp.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

/**
 * $Id: $
 */

package com.linkedin.restli.client;

import java.lang.reflect.Array;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.linkedin.data.DataList;
import com.linkedin.data.DataMap;
import com.linkedin.data.schema.PathSpec;
import com.linkedin.data.template.DataTemplate;
import com.linkedin.data.template.DataTemplateUtil;
import com.linkedin.jersey.api.uri.UriBuilder;
import com.linkedin.jersey.api.uri.UriTemplate;
import com.linkedin.restli.common.ComplexResourceKey;
import com.linkedin.restli.common.CompoundKey;
import com.linkedin.restli.common.ResourceSpec;
import com.linkedin.restli.common.RestConstants;
import com.linkedin.restli.internal.client.URIUtil;
import com.linkedin.restli.internal.common.QueryParamsDataMap;
import com.linkedin.restli.internal.common.URLEscaper;
import com.linkedin.restli.internal.common.URLEscaper.Escaping;
import com.linkedin.util.ArgumentUtil;

/**
 * @author Josh Walker
 * @version $Revision: $
 */

public abstract class AbstractRequestBuilder<K, V, R extends Request<?>> implements
    RequestBuilder<R>
{
  private final String                _baseURITemplate;
  protected final Map<String, String> _headers     = new HashMap<String, String>();
  protected final CompoundKey         _assocKey    = new CompoundKey();
  protected final Map<String, String> _pathKeys    = new HashMap<String, String>();
  protected final ResourceSpec        _resourceSpec;
  protected final DataMap             _queryParams = new DataMap();

  protected AbstractRequestBuilder(String baseURITemplate, ResourceSpec resourceSpec)
  {
    _baseURITemplate = baseURITemplate;
    _resourceSpec = resourceSpec;
  }

  public AbstractRequestBuilder<K, V, R> header(String key, String value)
  {
    addHeader(key, value);
    return this;
  }

  protected void addHeader(String key, String value)
  {
    _headers.put(key, value);
  }

  protected void addReqParam(String key, Object value)
  {
    ArgumentUtil.notNull(value, "value");
    addParam(key, value);
  }

  public AbstractRequestBuilder<K, V, R> pathKey(String name, Object value)
  {
    addPathKey(name, keyToString(value, Escaping.NO_ESCAPING));
    return this;
  }

  /*
   * Note that this method overrides the value at the given key, rather than adds to the collection
   * of values for it.
   */
  protected void addParam(String key, Object value)
  {
    if (value != null)
    {
      if (value instanceof Object[])
      {
        _queryParams.put(key, new DataList(stringify(Arrays.asList((Object[]) value))));
      }
      else if (value.getClass().isArray())
      { // not an array of objects but still an array - must be an array of primitives
        _queryParams.put(key, new DataList(stringify(value)));
      }
      else if (value instanceof DataTemplate)
      {
        @SuppressWarnings("rawtypes")
        DataTemplate dataTemplate = (DataTemplate)value;
        _queryParams.put(key, dataTemplate.data());
      }
      else if (value instanceof Iterable)
      {
        _queryParams.put(key, new DataList(stringify((Iterable<?>) value)));
      }
      else
      {
        _queryParams.put(key, stringifySimpleValue(value));
      }
    }
  }

  /**
   * To be called from the extending BatchXXXRequestBuilder classes that implement
   * ids(K...) or inputs()
   *
   * @param ids
   */
  protected final void addKeys(Collection<K> ids)
  {
    if (ids == null)
    {
      throw new IllegalArgumentException("null id");
    }

    Set<Object> allIds = new HashSet<Object>();
    for (K id : ids)
    {
      addKey(allIds, id);
    }

    DataList existingIds = (DataList)_queryParams.get(RestConstants.QUERY_BATCH_IDS_PARAM);
    if (existingIds != null && !existingIds.isEmpty())
    {
      allIds.addAll(existingIds);
    }

    _queryParams.put(RestConstants.QUERY_BATCH_IDS_PARAM,
                     new DataList(new ArrayList<Object>(allIds)));

  }

  /**
   * Adds the provided id to the provided set of objects that can be put into a DataMap (i.e. DataObject of String)
   * If the id is an instance of ComplexResourceKey, convert to DataMap, otherwise stringify.
   */
  private void addKey(Set<Object> ids, Object id)
  {
    if (id == null) {
      throw new IllegalArgumentException("Null key");
    }

    Object castId =
        id instanceof ComplexResourceKey<?, ?>
            ? ((ComplexResourceKey<?, ?>) id).toDataMap() : stringifySimpleValue(id);

    ids.add(castId);
  }

  /**
   * Add an individual key to the DataList of keys, which will be later resolved into a collection
   * of individual query parameters.
   */
  protected final void addKey(K id)
  {
    addKeys(Collections.singletonList(id));
  }

  /**
   * Depending on whether or not this is a complex resource key, return either a full (key + params)
   * string representation of the key (which for ComplexResource includes both key and params), or
   * simply key.toString() (which for ComplexResourceKey only uses key part).
   */
  protected String keyToString(K key)
  {
    return keyToString(key, Escaping.NO_ESCAPING);
  }

  protected void appendKeyToPath(UriBuilder uriBuilder, Object key)
  {
    uriBuilder.path(keyToString(key, Escaping.URL_ESCAPING));
  }

  private String keyToString(Object key, Escaping escaping)
  {
    if (key == null)
    {
      return null;
    }

    String result;
    if (key instanceof ComplexResourceKey)
    {
      result = ((ComplexResourceKey<?,?>)key).toStringFull(escaping);
    }
    else if (key instanceof CompoundKey)
    {
      result = key.toString(); // already escaped
    }
    else
    {
      result = URLEscaper.escape(stringifySimpleValue(key), escaping);
    }
    return result;
  }

  /** given a iterable of objects returns a list of (non-null) toStrings */
  private static List<String> stringify(Iterable<?> values)
  {
    assert values != null;
    List<String> strings =
        new ArrayList<String>(values instanceof Collection ? ((Collection<?>) values).size() : 10);
    for (Object value : values)
    {
      if (value != null)
      {
        strings.add(stringifySimpleValue(value));
      }
    }
    return strings;
  }

  static String stringifySimpleValue(Object value)
  {
    Class<?> valueClass = value.getClass();
    if (DataTemplateUtil.hasCoercer(valueClass))
    {
      @SuppressWarnings("unchecked")
      Class<Object> fromClass = (Class<Object>) value.getClass();
      return DataTemplateUtil.coerceInput(value, fromClass, Object.class).toString();
    }
    return value.toString();
  }

  /** given an array of primitives returns a collection of strings */
  private static List<String> stringify(Object array)
  {
    assert array != null && array.getClass().isArray();
    int len = Array.getLength(array);
    List<String> strings = new ArrayList<String>(len);
    for (int i = 0; i < len; ++i)
    {
      Object value = Array.get(array, i);
      if (value != null)
      {
        strings.add(value.toString());
      }
    }
    return strings;
  }

  protected void addPathKey(String key, Object value)
  {
    if (value == null)
    {
      throw new IllegalArgumentException("Path key must be non-null");
    }
    _pathKeys.put(key, stringifySimpleValue(value));
  }

  protected void addAssocKey(String key, Object value)
  {
    _assocKey.append(key, value);
  }

  protected void appendQueryParams(UriBuilder b)
  {
    QueryParamsDataMap.addSortedParams(b, _queryParams);
  }

  protected final void appendAssocKeys(UriBuilder uriBuilder)
  {
    uriBuilder.path(_assocKey.toString());
  }

  protected URI bindPathKeys()
  {
    UriTemplate template = new UriTemplate(_baseURITemplate);
    for (String key : template.getTemplateVariables())
    {
      if (!_pathKeys.containsKey(key))
      {
        throw new IllegalStateException("Missing path key: '" + key + "'");
      }
    }
    Map<String, String> escapedKeys = new HashMap<String, String>();
    for(Map.Entry<String, String> key : _pathKeys.entrySet())
    {
      escapedKeys.put(key.getKey(), URLEscaper.escape(key.getValue(), Escaping.URL_ESCAPING));
    }

    return URI.create(template.createURI(escapedKeys));
  }

  protected void addFields(PathSpec... fieldPaths)
  {
    if (_queryParams.containsKey(RestConstants.FIELDS_PARAM))
    {
      throw new IllegalStateException("Fields already set on this request: "
          + _queryParams.get(RestConstants.FIELDS_PARAM));
    }

    addParam(RestConstants.FIELDS_PARAM, URIUtil.encodeFields(fieldPaths));
  }

  protected boolean isComplexKeyResource()
  {
    return _resourceSpec.getKeyClass() == ComplexResourceKey.class;
  }

}
